%module strmat
%{
#include "stree.h"
%}


// This is a class wrapper over the strmat suffix trees library.
//
// Wrapper written by Danny Yoo (dyoo@hkn.eecs.berkeley.edu)
//
// Original code by Dan Gusfield's web site at:
//
//     http://www.cs.ucdavis.edu/~gusfield/strmat.html
//
// This is my first real SWIG wrapper, so please take the stability of
// the wrapper with a heaping spoonful of salt; it's not guaranteed to
// work right.


%include typemaps.i

%inline %{
  struct SuffixTree {
    STREE_STRUCT *tree;
  };

  struct SuffixNode {
    STREE_STRUCT *tree;
    SNODE_STRUCT *node;
  };

  typedef struct SuffixTree SuffixTree;
  typedef struct SuffixNode SuffixNode;

  %}

// Certain functions return garbage-collectable objects.  The following
// directives tell SWIG about this.
%newobject SuffixTree::SuffixTree;
%newobject SuffixTree::match;
%newobject SuffixTree::walk;
%newobject SuffixTree::root;


%apply int *OUTPUT { int *endpos_out };
%apply SWIGTYPE *OUTPUT { SuffixNode *node_out };


%addmethods SuffixTree {
  SuffixTree() {
    const int COPY_FLAG = 1;
    SuffixTree *self = malloc(sizeof(SuffixTree));
    self->tree = stree_new_tree(STREE_ASCII, COPY_FLAG);
    return self;
  };

  ~SuffixTree() {
    stree_delete_tree(self->tree);
    free(self);
  };

  int add_string(char* s, int id) {
      return stree_add_string(self->tree, s, strlen(s), id);
  };


//    I don't know if string removal is safe --- this appears to dump core.
//    I wonder why!
//    int remove_string(int id) {
//      return stree_remove_string(self->tree, id);
//    }

  int get42(int *endpos_out) {
    *endpos_out = 42;
    return 24;
  }

  PyObject *match2(char *s, SuffixNode *node_out, int *endpos_out) {
    STREE_NODE stree_node;
    int matched_chars = stree_match(self->tree, s, strlen(s), 
				    &stree_node, endpos_out);
    node_out->tree = self->tree;
    node_out->node = stree_node;
    return matched_chars;
  }



  // Note: the following is really really darn ugly.  I will make a
  // typemap as soon as I figure out how to generalize this SWIGTYPE
  // OUTPUT pattern.
#ifdef SWIGPYTHON
  PyObject *match(char *s) {
    STREE_NODE node_out;
    int endpos_out;
    int matched_chars;
    PyObject* tuple;
    SuffixNode* node;
    matched_chars = stree_match(self->tree, s, strlen(s),
				&node_out, &endpos_out);
    tuple = PyTuple_New(3);
    PyTuple_SetItem(tuple, 0, PyInt_FromLong((long)matched_chars));
    PyTuple_SetItem(tuple, 2, PyInt_FromLong((long)endpos_out));

    node = malloc(sizeof(SuffixTree));
    node->tree = self->tree;
    node->node = node_out;
    if (node->node == NULL) {
      free(node);
      PyTuple_SetItem(tuple, 1, Py_None);
      Py_INCREF(Py_None);
    }
    else {
      PyTuple_SetItem(tuple, 1, 
		      SWIG_NewPointerObj((void*)node,
					 SWIGTYPE_p_SuffixNode,
					 1));
    }
    return tuple;
  }
#endif

//    Not done yet
//    SuffixNode* walk() {
//      SuffixNode *result = malloc(sizeof(SuffixNode));
//      result->tree = self->tree;
//      return result; // testing; fixme, this is not right yet
//    }


  int num_nodes() {
    return stree_get_num_nodes(self->tree);
  };


  SuffixNode* root() {
    SuffixNode *s = malloc(sizeof(SuffixNode));
    s->tree = self->tree;
    s->node = stree_get_root(self->tree);
    if (s->node == NULL) {
      free(s);
      return NULL;
    }
    return s;
  };

};

//////////////////////////////////////////////////////////////////////

// Certain functions return garbage-collectable objects.  The following
// directives tell SWIG about this.
%newobject SuffixNode::find_child;
%newobject SuffixNode::children;
%newobject SuffixNode::next;
%newobject SuffixNode::parent;
%newobject SuffixNode::suffix_link;
%newobject SuffixNode::leaf;


// Also, some functions return allocated character strings, so these
// should also be under garbage collection.
%typemap(newfree) char* {
  printf("Freeing string %s!", $1);
  free($1);
}
%newobject SuffixNode::edgestr;
%newobject SuffixNode::labelstr;
%newobject SuffixNode::getch;


%addmethods SuffixNode {
  ~SuffixNode() {
    free(self);
  }


  int num_children() {
    return stree_get_num_children(self->tree, self->node);
  }


  // Testme!  How does find_child work?  What does it do?
  SuffixNode* find_child(char ch) {
    SuffixNode* child = malloc(sizeof(SuffixNode));
    child->tree = self->tree;
    child->node = stree_find_child(self->tree, self->node, ch);
    if (child->node == NULL) {
      free(child);
      return NULL;
    }
    return child;
  }


  // We may want to change this so that it returns a list of children, so that
  // we don't have to think about linked list stuff.
  SuffixNode* children() {
    SuffixNode* child = malloc(sizeof(SuffixNode));
    child->tree = self->tree;
    child->node = stree_get_children(self->tree, self->node);
    if (child->node == NULL) {
      free(child);
      return NULL;
    }
    return child;
  }


  SuffixNode* next() {
    SuffixNode* sibling = malloc(sizeof(SuffixNode));
    sibling->tree = self->tree;
    sibling->node = stree_get_next(self->tree, self->node);
    if (sibling->node == NULL) {
      free(sibling);
      return NULL;
    }
    return sibling;
  }


  SuffixNode* parent() {
    SuffixNode* parent = malloc(sizeof(SuffixNode));
    parent->tree = self->tree;
    parent->node = stree_get_parent(self->tree, self->node);
    if (parent->node == NULL) {
      free(parent);
      return NULL;
    }
    return parent;
  }


  SuffixNode* suffix_link() {
    SuffixNode* suffix = malloc(sizeof(SuffixNode));
    suffix->tree = self->tree;
    suffix->node = stree_get_suffix_link(self->tree, self->node);
    if (suffix->node == NULL) {
      free(suffix);
      return NULL;
    }
    return suffix;
  }


  int edgelen() {
    return stree_get_edgelen(self->tree, self->node);
  }


  char* edgestr() {
    char *result_str = NULL;
    int len = stree_get_edgelen(self->tree, self->node);
    result_str = malloc(sizeof(char) * (len+1));
    strncpy(result_str, stree_get_edgestr(self->tree, self->node), len);
    result_str[len] = '\0';
    return result_str;
  };


  char* getch() {
    /** If the node is a root, it has no 'ch', so we must take direct
	action to avoid a segfault.
	SWIG.**/
    char *result;
    if (self->node == stree_get_root(self->tree)) {
      return NULL;
    }
    result = malloc(sizeof(char) + 1);
    result[0] = stree_getch(self->tree, self->node); result[1] = '\0';
    return result;
  };
  

  int labellen() {
    return stree_get_labellen(self->tree, self->node);
  }


  char* labelstr() {
    char *result_str = NULL;
    int len = stree_get_labellen(self->tree, self->node);
    result_str = malloc(sizeof(char) * (len+1));
    stree_get_label(self->tree, self->node, result_str,
  		    len, 0);
    result_str[len] = '\0';   // Perhaps this isn't necessary...
    return result_str;
  } 


  int ident() {
    return stree_get_ident(self->tree, self->node);
  }

  int num_leaves() {
    return stree_get_num_leaves(self->tree, self->node);
  }


#ifdef SWIGPYTHON
  PyObject* leaf(int leafnum) {
    char *string_out;
    int pos_out;
    int id_out;
    PyObject *tuple;
    if(stree_get_leaf(self->tree, self->node, leafnum,
		      &string_out, &pos_out, &id_out)) {
      tuple = PyTuple_New(3);
      PyTuple_SetItem(tuple, 0, PyString_FromString(string_out));
      PyTuple_SetItem(tuple, 1, PyInt_FromLong((long)pos_out));
      PyTuple_SetItem(tuple, 2, PyInt_FromLong((long)id_out));
      return tuple;
    }
    else {
      Py_INCREF(Py_None);
      return Py_None;
    }
  }
#endif


  // Returns the id() we used to actually insert this string.  It's
  // very strange to me that the original strmat library doesn't have
  // this functionality!  I've also had to modify the stree.h to add
  // this method.
  int id() {
    return stree_get_strid(self->tree, self->node);
  }

};

%rename(set_max_alloc) stree_set_max_alloc;
void stree_set_max_alloc(int size);
